package cz.sa.shiftPlanner.servicesImpl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.dozer.Mapper;
import org.joda.time.DateTime;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import cz.sa.shiftPlanner.dao.BusConnectionDAO;
import cz.sa.shiftPlanner.dao.NotificationDAO;
import cz.sa.shiftPlanner.dao.PersonDAO;
import cz.sa.shiftPlanner.dto.BusConnectionDTO;
import cz.sa.shiftPlanner.dto.BusTimetableDTO;
import cz.sa.shiftPlanner.dto.BusTurnsTimetableDTO;
import cz.sa.shiftPlanner.dto.CityDTO;
import cz.sa.shiftPlanner.dto.LineGroupDTO;
import cz.sa.shiftPlanner.dto.StationDTO;
import cz.sa.shiftPlanner.entities.BusConnectionVO;
import cz.sa.shiftPlanner.entities.LineGroupVO;
import cz.sa.shiftPlanner.entities.NotificationVO;
import cz.sa.shiftPlanner.entities.StationVO;
import cz.sa.shiftPlanner.enums.NotifType;
import cz.sa.shiftPlanner.enums.WeekDay;
import cz.sa.shiftPlanner.services.BusConnectionService;
import cz.sa.shiftPlanner.services.BusTurnsTimetableService;
import cz.sa.shiftPlanner.utils.BusConnectionComparator;
import cz.sa.shiftPlanner.utils.BusConnectionListComparator;
import cz.sa.shiftPlanner.utils.DateFormatter;
import cz.sa.shiftPlanner.utils.TimeRange;
import cz.sa.shiftPlanner.utils.Tuple;

@Service
@Transactional(propagation = Propagation.REQUIRED)
public class BusConnectionServiceImpl implements BusConnectionService {

  private static final Logger log = Logger.getLogger(BusConnectionServiceImpl.class);
  
  private BusConnectionDAO busConnectionDAO;
  private BusTurnsTimetableService busTurnsTimetableService;
  private Mapper mapper;

  @Autowired
  @Qualifier("notificationDAO")
  private NotificationDAO notificationDAO;

  @Autowired
  @Qualifier("personDAO")
  private PersonDAO personDAO;

  public void setBusTurnsTimetableService(BusTurnsTimetableService busTurnsTimetableService) {
    this.busTurnsTimetableService = busTurnsTimetableService;
  }

  public Mapper getMapper() {
    return mapper;
  }

  public void setMapper(Mapper mapper) {
    this.mapper = mapper;
  }

  public BusConnectionDAO getBusConnectionDAO() {
    return busConnectionDAO;
  }

  public void setBusConnectionDAO(BusConnectionDAO busConnectionDAO) {
    this.busConnectionDAO = busConnectionDAO;
  }

  @Override
  @Transactional(readOnly = true)
  public BusConnectionDTO get(Long id) {
    return mapper.map(busConnectionDAO.get(id), BusConnectionDTO.class);
  }

  @Override
  @Transactional
  public void create(BusConnectionDTO connection) {
    busConnectionDAO.create(mapper.map(connection, BusConnectionVO.class));

  }

  @Override
  @Transactional
  public void update(BusConnectionDTO connection) {
    busConnectionDAO.update(mapper.map(connection, BusConnectionVO.class));

  }

  @Override
  @Transactional
  public void remove(BusConnectionDTO connection) {
    busConnectionDAO.remove(mapper.map(connection, BusConnectionVO.class));

  }

  @Override
  @Transactional(readOnly = true)
  public List<BusConnectionDTO> findAll() {
    log.info("find all begin");
    List<BusConnectionDTO> list = new ArrayList<BusConnectionDTO>();
    for (BusConnectionVO vo : busConnectionDAO.findAll())
      list.add(mapper.map(vo, BusConnectionDTO.class));
    return list;
  }

  /**
   * return connections between arrival and departure
   */
  @Override
  @Transactional(readOnly = true)
  public List<BusConnectionDTO> findConnectionByTime(DateTime departure, DateTime arrival) {
    log.info("findConnectionByTime begin");
    List<BusConnectionDTO> list = new ArrayList<BusConnectionDTO>();
    int i = 0;
    for (BusConnectionVO vo : busConnectionDAO.findConnectionByTime(departure, arrival)){
      list.add(mapper.map(vo, BusConnectionDTO.class));
      System.out.println(i++);
    }
    log.info("findConnectionByTime end");
    return list;
  }

  @Override
  @Transactional(readOnly = true)
  public BusConnectionDTO findConnectionByConcreteFigure(DateTime departure, DateTime arrival, StationDTO departureCity,
      StationDTO arrivalCity, LineGroupDTO line) {

    BusConnectionVO vo = busConnectionDAO.findConnectionByConcreteFigure(departure, arrival,
        mapper.map(departureCity, StationVO.class), mapper.map(arrivalCity, StationVO.class), mapper.map(line, LineGroupVO.class));

    if (vo == null)
      return null;
    return mapper.map(vo, BusConnectionDTO.class);

  }

  @Override
  @Transactional(readOnly = true)
  public List<List<BusConnectionDTO>> findConnectionByTimeRange(TimeRange range, LineGroupDTO lg) {

    List<BusConnectionDTO> list = new ArrayList<BusConnectionDTO>();
    for (BusConnectionVO vo : busConnectionDAO.findConnectionByTimeRange(range, mapper.map(lg, LineGroupVO.class))) {
      list.add(mapper.map(vo, BusConnectionDTO.class));
    }

    if (list.size() == 0)
      return new ArrayList<List<BusConnectionDTO>>();
    List<WeekDay> days = Arrays.asList(WeekDay.values());
    List<List<BusConnectionDTO>> sortedField = sortListToLists(list, days.get(range.getLower().getDayOfWeek() - 1),range.getLower());
    Collections.sort(sortedField, new BusConnectionListComparator());
    sortedField = sortToCouples(sortedField);
    return sortedField;
  }

  /**
   * tahle agenda je kvuli slozenym linkam, kde je vice nez dve vyjezdni mesta. 
   * Prijima seznam kde jsou prvky serazeny podle casu a dela skupinky podle vyjezdnich a prijezdovych mest.
   * @param sortingField
   * @return List<List<BusConnectionDTO>
   */
  private List<List<BusConnectionDTO>> sortToCouples(List<List<BusConnectionDTO>> sortingField) {
    Map<Tuple<CityDTO, CityDTO>, List<List<BusConnectionDTO>>> map = new LinkedHashMap<Tuple<CityDTO, CityDTO>, List<List<BusConnectionDTO>>>();
    
    for (List<BusConnectionDTO> list : sortingField) {
      Tuple<CityDTO, CityDTO> key = new Tuple<CityDTO, CityDTO>(
          list.get(0).getTimetable().getDepartureStation().getCity(),
          list.get(0).getTimetable().getArrivalStation().getCity());
      if(map.containsKey(key)){ 
        List<List<BusConnectionDTO>> value = map.get(key);
        if(value.contains(list)) continue;
        value.add(list);
        map.put(key, value);
      }else{
        List<List<BusConnectionDTO>> value = new ArrayList<List<BusConnectionDTO>>();
        value.add(list);
        map.put(key, value);
        //pokusim se hned vytvorit opacny smer aby byly smery tam z5 ve vypisu pohromade
        key = new Tuple<CityDTO, CityDTO>(
            list.get(0).getTimetable().getArrivalStation().getCity(),
            list.get(0).getTimetable().getDepartureStation().getCity());
        if(!map.containsKey(key)){ 
          value = new ArrayList<List<BusConnectionDTO>>();
          map.put(key, value);
        }
      }
    }
    List<List<BusConnectionDTO>> result = new ArrayList<List<BusConnectionDTO>>();
    for (List<List<BusConnectionDTO>>  item : map.values()) {
      result.addAll(item);
    }
    return result;
  }

  /**
   * Radi BusConnections podle couple number do jednotlivych tocek
   * @param list
   * @param day
   * @param validity
   * @return seznam seznamu tocek
   */
  private List<List<BusConnectionDTO>> sortListToLists(List<BusConnectionDTO> list, WeekDay day, DateTime validity) {
    List<List<BusConnectionDTO>> result = new ArrayList<List<BusConnectionDTO>>();
    if (list.size() == 0)
      return result;

    BusTurnsTimetableDTO b = busTurnsTimetableService.findByDayAndTimetable(list.get(0).getTimetable(), day, validity,
        DateFormatter.getWeekMod2(validity));
    if (b == null)
      return result;
    int coupleNumber = b.getCoupleNumber();

    List<BusConnectionDTO> listLoc = new ArrayList<BusConnectionDTO>();

    //cyklus projde vsechny zaznamy a rozseka je podle dnu
    for (BusConnectionDTO dto : list) {
      //pokud se objevi bez couple number nebo bez Turn. Pak to znamena chybu a je treba tento spoj nahlasit 
      BusTurnsTimetableDTO turn = busTurnsTimetableService.findByDayAndTimetable(dto.getTimetable(), day, validity,
          DateFormatter.getWeekMod2(validity));
      Integer coupleNumberLoc = turn == null ? null : turn.getCoupleNumber();
      if (coupleNumberLoc == null)
        throw new IllegalArgumentException("problem with couplenumber");
      dto.setBusTurn(turn);
      if (coupleNumber == coupleNumberLoc) {
        if (!listLoc.contains(dto))
          listLoc.add(dto);
      }
      else {
        result.add(listLoc);
        coupleNumber = coupleNumberLoc;
        listLoc = new ArrayList<BusConnectionDTO>();
        listLoc.add(dto);
      }
    }
    if (!result.contains(listLoc))
      result.add(listLoc);
    return result;
  }


  @Override
  public void checkMorning() {

    DateTime actual = new DateTime();
    DateTime nextDayFrom;
    DateTime nextDayTo;
    Calendar cal = Calendar.getInstance();
    int maxDay = cal.getActualMaximum(Calendar.DAY_OF_MONTH);
    if (actual.getDayOfMonth() == maxDay) {
      nextDayFrom = new DateTime(actual.getYear(), actual.getMonthOfYear() + 1, 1, 0, 0, 0, 1);
      nextDayTo = new DateTime(actual.getYear(), actual.getMonthOfYear() + 1, 1, 14, 0, 0, 0);
    }
    else {
      nextDayFrom = new DateTime(actual.getYear(), actual.getMonthOfYear(), actual.getDayOfMonth() + 1, 0, 0, 0, 1);
      nextDayTo = new DateTime(actual.getYear(), actual.getMonthOfYear(), actual.getDayOfMonth() + 1, 14, 0, 0, 0);
    }
    List<BusConnectionDTO> list = findConnectionByTime(nextDayFrom, nextDayTo);
    for (BusConnectionDTO c : list) {
      NotificationVO note = new NotificationVO();
      note.setCreated(new DateTime());
      note.setPerson(personDAO.findByPersonalNumber(4L));
      note.setReaded(false);
      note.setNotifType(NotifType.KONTROLY);
      note.setText("Spoj z " + c.getTimetable().getDepartureStation().getCity() + " " + DateFormatter.toDateAndTime(c.getDepartureTime())
          + " do " + c.getTimetable().getArrivalStation().getCity() + " " + DateFormatter.toDateAndTime(c.getArrivalTime())
          + " neni obsazen");
      notificationDAO.create(note);
    }
  }

  @Override
  public void checkAfternoon() {

    DateTime actual = new DateTime();
    DateTime nextDayFrom;
    DateTime nextDayTo;
    Calendar cal = Calendar.getInstance();
    int maxDay = cal.getActualMaximum(Calendar.DAY_OF_MONTH);
    if (actual.getDayOfMonth() == maxDay) {
      nextDayFrom = new DateTime(actual.getYear(), actual.getMonthOfYear() + 1, 1, 12, 0, 0, 0);
      nextDayTo = new DateTime(actual.getYear(), actual.getMonthOfYear() + 1, 2, 2, 0, 0, 0);
    }
    else {
      nextDayFrom = new DateTime(actual.getYear(), actual.getMonthOfYear(), actual.getDayOfMonth() + 1, 12, 0, 0, 0);
      nextDayTo = new DateTime(actual.getYear(), actual.getMonthOfYear(), actual.getDayOfMonth() + 2, 2, 0, 0, 0);
    }
    List<BusConnectionDTO> list = findConnectionByTime(nextDayFrom, nextDayTo);
    for (BusConnectionDTO c : list) {
      NotificationVO note = new NotificationVO();
      note.setCreated(new DateTime());
      note.setPerson(personDAO.findByPersonalNumber(4L));
      note.setReaded(false);
      note.setNotifType(NotifType.KONTROLY);
      note.setText("Spoj z " + c.getTimetable().getDepartureStation().getCity() + " " + DateFormatter.toDateAndTime(c.getDepartureTime())
          + " do " + c.getTimetable().getArrivalStation().getCity() + " " + DateFormatter.toDateAndTime(c.getArrivalTime())
          + " neni obsazen");
      notificationDAO.create(note);
    }
  }

  @Override
  public List<BusConnectionDTO> findSupportsByTimeRange(DateTime day, Long groupLine) {

    List<BusConnectionDTO> list = new ArrayList<BusConnectionDTO>();
    for (BusConnectionVO vo : busConnectionDAO.findSupportsByTimeRange(day, groupLine)) {
      list.add(mapper.map(vo, BusConnectionDTO.class));
    }
    Collections.sort(list, new BusConnectionComparator());
    return list;
  }
  
  @Override
  public BusConnectionDTO findFirstEmpty(LineGroupDTO lg) {
    BusConnectionVO conn = busConnectionDAO.findFirstEmpty(mapper.map(lg, LineGroupVO.class));
    if(conn == null)
      return null;
    else
      return mapper.map(conn, BusConnectionDTO.class);
  }

}
