package cz.sa.shiftPlanner.utils;

import java.util.Arrays;
import java.util.Date;
import java.util.List;

import javax.mail.internet.MimeMessage;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.task.TaskExecutor;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.mail.javamail.MimeMessageHelper;
import org.springframework.mail.javamail.MimeMessagePreparator;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.support.TransactionSynchronization;
import org.springframework.transaction.support.TransactionSynchronizationAdapter;
import org.springframework.transaction.support.TransactionSynchronizationManager;
import org.springframework.util.StringUtils;

@Service
public class Mailer {

  private static final String MAIL_ENCODING = "UTF-8";

  private final Logger log = LoggerFactory.getLogger(getClass());

  @Autowired
  private Config config;
  @Autowired
  private TaskExecutor taskExecutor;
  @Autowired
  private JavaMailSender mailSender;

  @Transactional(propagation = Propagation.MANDATORY)
  public void sendMailAfterCommit(final String from, final String to, final String subject, final String plainText,
      final String htmlText) {

    if (!StringUtils.hasText(to)) {
      log.debug("Empty email address. Subject: " + subject);
      return;
    }

    // plainText musi byt zadan vzdy. htmlText je volitelny.
    if (plainText == null) {
      log.warn("Empty body. Could not send email to " + to);
      return;
    }

    TransactionSynchronization synchronization = new TransactionSynchronizationAdapter() {
      @Override
      public void afterCommit() {
        sendMultiMail(from, to, subject, plainText, htmlText);
      }

    };

    TransactionSynchronizationManager.registerSynchronization(synchronization);
  }

  private String reparirAddress(String str) {
    return str.replaceAll("[ \t\n\r]", "");
  }

  /**
   * Zaslani emailu na vice adres.
   * @param from Email odesilatele
   * @param toEmails Emaily prijemcu oddelene carkou ci strednikem
   * @param subject Predmet zpravy
   * @param plainText Text zpravy - plaintext
   * @param htmlText Text zpravy - HTML
   */
  public void sendMultiMail(final String from, String toEmails, final String subject, final String plainText, final String htmlText) {
    String[] toEmailsArray = StringUtils.hasText(toEmails) ? toEmails.split("\\s*[,;]\\s*") : new String[] { toEmails };
    for (String to : toEmailsArray) {
      to = StringUtils.trimWhitespace(to);
      sendMail(from, to, subject, plainText, htmlText);
    }
  }
  
  /**
   * Zaslani emailu na jedinou emailovou adresu.
   * @param from Email odesilatele
   * @param to Jediny email prijemce
   * @param subject Predmet zpravy
   * @param plainText Text zpravy - plaintext
   * @param htmlText Text zpravy - HTML
   */
  public void sendMail(final String from, String to, final String subject, final String plainText, final String htmlText){
    //abychom pri vyvoji omylem neposilali maily realnym zakaznikum
    if (!canSendMail(to)) { return; }

    final String repariredTo = reparirAddress(to);
    taskExecutor.execute(new Runnable() {
      public void run() {
        try {
          mailSender.send(new MimeMessagePreparator() {
            public void prepare(MimeMessage mimeMessage) throws Exception {
              MimeMessageHelper message = new MimeMessageHelper(mimeMessage, true, MAIL_ENCODING);
              message.setSentDate(new Date());
              message.setFrom(from);
              message.setTo(repariredTo);
              message.setSubject(subject);

              if (plainText != null && htmlText != null) {
                message.setText(plainText, htmlText);
              } else if (plainText != null) {
                message.setText(plainText, false);
              } else if (htmlText != null) {
                message.setText(htmlText, true);
              } else {
                throw new IllegalStateException();
              }
            }
          });
          log.info("Mail send successfuly to: " + repariredTo + ", from: " + from + ", subject: " + subject);
        } catch (Exception e) {
          log.error("Error sending mail to: " + repariredTo, e);
        }
      }
    });
  }
  

  private boolean canSendMail(String to) {
    String skippingMailMessageBase = "Skipping email to " + to + " : ";

    if (!StringUtils.hasText(to)) {
      log.info(skippingMailMessageBase + "'to' is empty");
      return false;
    }

    if (!StringUtils.hasText(config.getProperty("smtp.server"))) {
      log.info(skippingMailMessageBase + "smtp.server is empty");
      return false;
    }

    boolean sendAllEmails = "true".equals(config.getProperty("sendAllEmails"));
    if (!sendAllEmails && !getDevelopementEmails().contains(to.toLowerCase())) {
      log.info(skippingMailMessageBase + "sendAllEmails=false");
      return false;
    }

    return true;
  }

  /**
   * emaily developeru, nastavuji se v plannerWeb.poperties
   * 
   * @return
   */
  public List<String> getDevelopementEmails() {
    return Arrays.asList(config.getProperty("developersEmails").split("(\\s*,\\s*)|(\\s*;\\s*)"));
  }

}
